/* latency.c -- measure latency of OS */

#include "porttime.h"
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "assert.h"

/* Latency is defined here to mean the time starting when a
   process becomes ready to run, and ending when the process
   actually runs. Latency is due to contention for the
   processor, usually due to other processes, OS activity
   including device drivers handling interrupts, and
   waiting for the scheduler to suspend the currently running
   process and activate the one that is waiting.

   Latency can affect PortMidi applications: if a process fails
   to wake up promptly, MIDI input may sit in the input buffer
   waiting to be handled, and MIDI output may not be generated
   with accurate timing. Using the latency parameter when 
   opening a MIDI output port allows the caller to defer timing
   to PortMidi, which in most implementations will pass the
   data on to the OS. By passing timestamps and data to the
   OS kernel, device driver, or even hardware, there are fewer
   sources of latency that can affect the ultimate timing of
   the data. On the other hand, the application must generate
   and deliver the data ahead of the timestamp. The amount by 
   which data is computed early must be at least as large as
   the worst-case latency to avoid timing problems.

   Latency is even more important in audio applications. If an
   application lets an audio output buffer underflow, an audible
   pop or click is produced. Audio input buffers can overflow,
   causing data to be lost. In general the audio buffers must
   be large enough to buffer the worst-case latency that the
   application will encounter.

   This program measures latency by recording the difference
   between the scheduled callback time and the current real time.
   We do not really know the scheduled callback time, so we will
   record the differences between the real time of each callback
   and the real time of the previous callback. Differences that
   are larger than the scheduled difference are recorded. Smaller
   differences indicate the system is recovering from an earlier
   latency, so these are ignored.
   Since printing by the callback process can cause all sorts of
   delays, this program records latency observations in a
   histogram. When the program is stopped, the histogram is
   printed to the console.

   INTERPRETING RESULTS: Time is quantized to 1ms, so there is
   some uncertainty due to rounding. A microsecond latency that
   spans the time when the clock is incremented will be reported
   as a latency of 1. On the other hand, a latency of almost
   1ms that falls between two clock ticks will be reported as 
   zero. In general, if the highest nonzero bin is numbered N,
   then the maximum latency is N+1.
 */

#define HIST_LEN 21 /* how many 1ms bins in the histogram */
#define PERIOD 2 /* milliseconds per callback */

#define STRING_MAX 80 /* used for console input */

#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) <= (b) ? (a) : (b))

PtTimestamp previous_callback_time = 0;

long histogram[HIST_LEN];
long max_latency = 0;  /* worst latency observed */
long out_of_range = 0; /* how many points outside of HIST_LEN? */

/* callback function for PortTime -- computes histogram */
void pt_callback(PtTimestamp timestamp, void *userData)
{
	PtTimestamp difference = timestamp - previous_callback_time - PERIOD;
	previous_callback_time = timestamp;

    /* allow 5 seconds for the system to settle down */
	if (timestamp < 5000) return;
	if (difference < 0) return; /* ignore when system is "catching up" */

	/* update the histogram */
	if (difference < HIST_LEN) {
		histogram[difference]++;
	} else {
		out_of_range++;
	}

	if (max_latency < difference) max_latency = difference;
}


int main()
{
	char line[STRING_MAX];
	int i;
	int len;
	PtTimestamp stop;
	printf("Latency histogram. Timer period = %dms\n", PERIOD);
	printf("%s%s", "Latency measurements will start in 5 seconds. ",
		           "Type return to stop: ");
	Pt_Start(PERIOD, &pt_callback, 0);
    fgets(line, STRING_MAX, stdin);
    stop = Pt_Time();
	Pt_Stop();

	/* print the histogram */
	printf("Duration of test: %g seconds\n\n", max(0, stop - 5000) * 0.001);
	printf("Latency(ms)  Number of occurrences\n");
	/* avoid printing beyond last non-zero histogram entry */
	len = min(HIST_LEN, max_latency + 1);
	for (i = 0; i < len; i++) {
		printf("%2d      %10d\n", i, histogram[i]);
	}
	printf("Number of points greater than %dms: %d\n", 
		   HIST_LEN - 1, out_of_range);
	printf("Maximum latency: %d milliseconds\n", max_latency);
	printf("\nNote that due to rounding, actual latency can be 1ms higher\n");
	printf("than the numbers reported here.\n");
	printf("Type return to exit...");
	fgets(line, STRING_MAX, stdin);
	return 0;
}

